from __future__ import annotations

from typing import TYPE_CHECKING

from framesss.pre.cases import LoadCombination

if TYPE_CHECKING:
    from desssign.loads.load_case import DesignLoadCase


class DesignLoadCombination(LoadCombination):
    def __init__(
        self,
        label: str,
        load_cases: dict[DesignLoadCase, float],
        combination_key: str,
    ) -> None:
        super().__init__(label, load_cases)  # type: ignore[arg-type]
        self.combination_key = combination_key



from __future__ import annotations

from itertools import combinations
from typing import TYPE_CHECKING

from desssign.loads.enums import LoadCaseRelation

if TYPE_CHECKING:
    from desssign.loads.load_case import DesignLoadCase


class LoadGroup:
    def __init__(
        self,
        load_cases: list[DesignLoadCase],
        load_case_relation: str | LoadCaseRelation = LoadCaseRelation.STANDARD,
    ) -> None:
        """
        Init the LoadGroup class.

        :param load_cases: List of :class:`DesignLoadCase` instances.
        :param load_case_relation: Relation between :class:`DesignLoadCase` instances.
                                   Possible values are ['together', 'standard', 'exclusive'].
        """
        self.load_cases = load_cases
        self.load_case_relation = LoadCaseRelation(load_case_relation)

    def __repr__(self) -> str:
        """Return a string representation of the LoadGroup object."""
        return (
            f"LoadGroup("
            f"load_cases={self.load_cases}, "
            f"load_case_relation={self.load_case_relation}"
            f")"
        )

    @property
    def number_of_load_cases(self) -> int:
        """Return the number of load cases."""
        return len(self.load_cases)

    @property
    def combinations(self) -> list[list[DesignLoadCase]]:
        """
        Generates a list of load case combinations based on the relation between load cases.

        :return: A list of lists, where each inner list represents a specific combination of `DesignLoadCase` instances.
        :raises AttributeError: If `load_case_relation` is not valid.
        """
        if self.load_case_relation == LoadCaseRelation.TOGETHER:
            return [self.load_cases]
        if self.load_case_relation == LoadCaseRelation.STANDARD:
            return [
                list(comb)
                for i in range(self.number_of_load_cases + 1)
                for comb in combinations(self.load_cases, i)
            ]
        if self.load_case_relation == LoadCaseRelation.EXCLUSIVE:
            comb = [[comb] for comb in self.load_cases]
            comb.insert(0, [])
            return comb
        else:
            raise AttributeError(
                f"Invalid load case relation: {self.load_case_relation}"
            )



from __future__ import annotations

from framesss.pre.cases import LoadCase

from desssign.loads.enums import LoadType
from desssign.loads.enums import VariableCategory


class DesignLoadCase(LoadCase):
    def __init__(
        self,
        label: str,
        load_type: str | LoadType,
        category: str | VariableCategory | None = None,
    ) -> None:
        """Init the DesignLoadCase class."""
        super().__init__(label)
        self.load_type = LoadType(load_type)

        if self.load_type == LoadType.VARIABLE:
            self.category: VariableCategory | None = VariableCategory(category)
        else:
            self.category = None

    def __repr__(self) -> str:
        """Return a string representation of the DesignLoadCase object."""
        # return (f"{self.__class__.__name__}("
        #         f"label={self.label}, "
        #         f"load_type={self.load_type}, "
        #         f"category={self.category})")
        return self.label



from __future__ import annotations

from itertools import product
from typing import TYPE_CHECKING

from desssign.loads.enums import LimitState
from desssign.loads.enums import LoadType
from desssign.loads.enums import SLSCombination
from desssign.loads.enums import ULSCombination
from desssign.loads.load_combination_generator.generate_combinations import (
    generate_combination,
)
from desssign.utils import flatten_list

if TYPE_CHECKING:
    from desssign.loads.load_combination import DesignLoadCombination
    from desssign.loads.load_group import LoadGroup


class CombinationGroup:
    def __init__(
        self,
        limit_state: str | LimitState,
        combination_type: str | SLSCombination | ULSCombination,
    ) -> None:
        """Init the CombinationGroup class."""
        self.limit_state = LimitState(limit_state)

        if self.limit_state == LimitState.ULS:
            self.combination_type = ULSCombination(combination_type)
        elif self.limit_state == LimitState.SLS:
            self.combination_type = SLSCombination(combination_type)
        else:
            raise AttributeError(
                f"Can't set combination type: '{combination_type}' to limit state: '{limit_state}'."
            )

        self.combinations: list[DesignLoadCombination] = []

    def generate_combinations(self, *args: list[LoadGroup]) -> None:
        # get all possible combinations
        all_iterables = []
        for load_groups in args:
            all_iterables.append(
                [load_group.combinations for load_group in load_groups]
            )

        combinations = []
        for iterables in all_iterables:
            combinations.extend(
                [flatten_list(combination) for combination in product(*iterables)]
            )

        # get all possible unique combinations
        unique_combinations = []
        for combination in combinations:
            if combination not in unique_combinations:
                unique_combinations.append(combination)

        c = 0
        for unique_combination in unique_combinations:
            permanent_cases = [
                case
                for case in unique_combination
                if case.load_type == LoadType.PERMANENT
            ]
            variable_cases = [
                case
                for case in unique_combination
                if case.load_type == LoadType.VARIABLE
            ]

            if variable_cases:  # in case there is only permanent cases
                for i, leading_variable_case in enumerate(variable_cases):
                    # loop through every possible combination of leading + other variable for this unique combination
                    other_variable_cases = variable_cases[:i] + variable_cases[i + 1 :]

                    self.combinations.extend(
                        generate_combination(
                            c,
                            permanent_cases,
                            leading_variable_case,
                            other_variable_cases,
                            self.combination_type,
                        )
                    )

                    c += 1
            else:
                self.combinations.extend(
                    generate_combination(
                        c, permanent_cases, None, [], self.combination_type
                    )
                )
                c += 1


if __name__ == "__main__":
    from desssign.loads.load_case import DesignLoadCase
    from desssign.loads.load_group import LoadGroup

    G1 = DesignLoadCase("G1", "permanent")
    G2 = DesignLoadCase("G2", "permanent")
    LG1 = LoadGroup([G1, G2], "together")

    Q1 = DesignLoadCase("Q1", "variable", "category a")
    Q2 = DesignLoadCase("Q2", "variable", "category b")
    LG2 = LoadGroup([Q1, Q2], "standard")

    S1 = DesignLoadCase("S1", "variable", "snow < 1000 m")
    S2 = DesignLoadCase("S2", "variable", "snow < 1000 m")
    S3 = DesignLoadCase("S3", "variable", "snow < 1000 m")
    LG3 = LoadGroup([S1, S2, S3], "exclusive")

    W1 = DesignLoadCase("W1", "variable", "wind")
    W2 = DesignLoadCase("W2", "variable", "wind")
    W3 = DesignLoadCase("W3", "variable", "wind")
    W4 = DesignLoadCase("W4", "variable", "wind")
    LG4 = LoadGroup([W1, W2, W3, W4], "exclusive")

    ULS = CombinationGroup("ULS", "basic")

    ULS.generate_combinations([LG1, LG2], [LG1, LG3, LG4])

    print("")



from __future__ import annotations

from typing import TYPE_CHECKING

from desssign.loads.enums import LoadBehavior
from desssign.loads.enums import SLSCombination
from desssign.loads.enums import ULSCombination
from desssign.loads.enums import VariableCategory
from desssign.loads.load_combination import DesignLoadCombination
from desssign.loads.load_combination_generator.constants import GAMMA_VALUES
from desssign.loads.load_combination_generator.constants import PSI_FACTORS
from desssign.loads.load_combination_generator.constants import XI

if TYPE_CHECKING:
    from desssign.loads.load_case import DesignLoadCase


def generate_combination(
    i: int,
    permanent_cases: list[DesignLoadCase],
    leading_variable_case: DesignLoadCase | None,
    other_variable_cases: list[DesignLoadCase],
    combination: SLSCombination | ULSCombination,
) -> list[DesignLoadCombination]:
    if combination == SLSCombination.CHARACTERISTIC:
        return generate_sls_characteristic_combination(
            i, permanent_cases, leading_variable_case, other_variable_cases
        )

    if combination == SLSCombination.FREQUENT:
        return generate_sls_frequent_combination(
            i, permanent_cases, leading_variable_case, other_variable_cases
        )

    if combination == SLSCombination.QUASIPERMANENT:
        return generate_sls_quasipermanent_combination(
            i, permanent_cases, leading_variable_case, other_variable_cases
        )

    if combination == ULSCombination.BASIC:
        return generate_uls_basic_combination(
            i, permanent_cases, leading_variable_case, other_variable_cases
        )

    if combination == ULSCombination.ALTERNATIVE:
        return generate_uls_alternative_combinations(
            i, permanent_cases, leading_variable_case, other_variable_cases
        )
    else:
        raise AttributeError(f"Unknown combination: '{combination}'.")


def generate_sls_characteristic_combination(
    i: int,
    permanent_cases: list[DesignLoadCase],
    leading_variable_case: DesignLoadCase | None,
    other_variable_cases: list[DesignLoadCase],
) -> list[DesignLoadCombination]:
    cases = {}
    key = ""

    for case in permanent_cases:
        cases[case] = 1.0
        key += f"{case.label}+"

    if leading_variable_case is not None:
        cases[leading_variable_case] = 1.0
        key += f"{leading_variable_case.label}+"

    for case in other_variable_cases:
        factor = PSI_FACTORS[VariableCategory(case.category)]["psi_0"]
        cases[case] = factor
        key += f"{factor}*{case.label}+"

    return [DesignLoadCombination(f"SLS-Characteristic-{i}", cases, key[:-1])]


def generate_sls_frequent_combination(
    i: int,
    permanent_cases: list[DesignLoadCase],
    leading_variable_case: DesignLoadCase | None,
    other_variable_cases: list[DesignLoadCase],
) -> list[DesignLoadCombination]:
    cases = {}
    key = ""

    for case in permanent_cases:
        cases[case] = 1.0
        key += f"{case.label}+"

    if leading_variable_case is not None:
        factor = PSI_FACTORS[VariableCategory(leading_variable_case.category)]["psi_1"]
        cases[leading_variable_case] = factor
        key += f"{factor}*{leading_variable_case.label}+"

    for case in other_variable_cases:
        factor = PSI_FACTORS[VariableCategory(case.category)]["psi_2"]
        cases[case] = factor
        key += f"{factor}*{case.label}+"

    return [DesignLoadCombination(f"SLS-Frequent-{i}", cases, key[:-1])]


def generate_sls_quasipermanent_combination(
    i: int,
    permanent_cases: list[DesignLoadCase],
    leading_variable_case: DesignLoadCase | None,
    other_variable_cases: list[DesignLoadCase],
) -> list[DesignLoadCombination]:
    cases = {}
    key = ""

    for case in permanent_cases:
        cases[case] = 1.0
        key += f"{case.label}+"

    if leading_variable_case is not None:
        factor = PSI_FACTORS[VariableCategory(leading_variable_case.category)]["psi_2"]
        cases[leading_variable_case] = factor
        key += f"{factor}*{leading_variable_case.label}+"

    for case in other_variable_cases:
        factor = PSI_FACTORS[VariableCategory(case.category)]["psi_2"]
        cases[case] = factor
        key += f"{factor}*{case.label}+"

    return [DesignLoadCombination(f"SLS-Frequent-{i}", cases, key[:-1])]


def generate_uls_basic_combination(
    i: int,
    permanent_cases: list[DesignLoadCase],
    leading_variable_case: DesignLoadCase | None,
    other_variable_cases: list[DesignLoadCase],
) -> list[DesignLoadCombination]:
    cases = {}
    key = ""

    for case in permanent_cases:
        factor = GAMMA_VALUES["Set B"][case.load_type][
            LoadBehavior.UNFAVOURABLE
        ]  # TODO: Favourable?
        cases[case] = factor
        key += f"{factor}*{case.label}+"

    if leading_variable_case is not None:
        factor = GAMMA_VALUES["Set B"][leading_variable_case.load_type][
            LoadBehavior.UNFAVOURABLE
        ]  # TODO: Favourable?
        cases[leading_variable_case] = factor
        key += f"{factor}*{leading_variable_case.label}+"

    for case in other_variable_cases:
        gamma = GAMMA_VALUES["Set B"][case.load_type][
            LoadBehavior.UNFAVOURABLE
        ]  # TODO: Favourable?
        psi = PSI_FACTORS[VariableCategory(case.category)]["psi_0"]
        cases[case] = gamma * psi
        key += f"{gamma}*{psi}*{case.label}+"

    return [DesignLoadCombination(f"ULS-Basic(6.10)-{i}", cases, key[:-1])]


def generate_uls_alternative_combinations(
    i: int,
    permanent_cases: list[DesignLoadCase],
    leading_variable_case: DesignLoadCase | None,
    other_variable_cases: list[DesignLoadCase],
) -> list[DesignLoadCombination]:
    cases_a = {}
    key_a = ""

    for case in permanent_cases:
        gamma = GAMMA_VALUES["Set B"][case.load_type][
            LoadBehavior.UNFAVOURABLE
        ]  # TODO: Favourable?
        cases_a[case] = gamma
        key_a += f"{gamma}*{case.label}+"

    if leading_variable_case is not None:
        gamma = GAMMA_VALUES["Set B"][leading_variable_case.load_type][
            LoadBehavior.UNFAVOURABLE
        ]  # TODO: Favourable?
        psi = PSI_FACTORS[VariableCategory(leading_variable_case.category)]["psi_0"]
        cases_a[leading_variable_case] = gamma * psi
        key_a += f"{gamma}*{psi}*{leading_variable_case.label}+"

    for case in other_variable_cases:
        gamma = GAMMA_VALUES["Set B"][case.load_type][
            LoadBehavior.UNFAVOURABLE
        ]  # TODO: Favourable?
        psi = PSI_FACTORS[VariableCategory(case.category)]["psi_0"]
        cases_a[case] = gamma * psi
        key_a += f"{gamma}*{psi}*{case.label}+"

    cases_b = {}
    key_b = ""

    for case in permanent_cases:
        gamma = GAMMA_VALUES["Set B"][case.load_type][
            LoadBehavior.UNFAVOURABLE
        ]  # TODO: Favourable?
        cases_b[case] = gamma * XI
        key_b += f"{XI}*{gamma}*{case.label}+"

    if leading_variable_case is not None:
        gamma = GAMMA_VALUES["Set B"][leading_variable_case.load_type][
            LoadBehavior.UNFAVOURABLE
        ]  # TODO: Favourable?
        cases_b[leading_variable_case] = gamma
        key_b += f"{gamma}*{leading_variable_case.label}+"

    for case in other_variable_cases:
        gamma = GAMMA_VALUES["Set B"][case.load_type][
            LoadBehavior.UNFAVOURABLE
        ]  # TODO: Favourable?
        psi = PSI_FACTORS[VariableCategory(case.category)]["psi_0"]
        cases_b[case] = gamma * psi
        key_b += f"{gamma}*{psi}*{case.label}+"

    combinations = [
        DesignLoadCombination(f"ULS-Alternative(6.10a)-{i}", cases_a, key_a[:-1]),
        DesignLoadCombination(f"ULS-Alternative(6.10b)-{i}", cases_b, key_b[:-1]),
    ]

    return combinations
